from pydantic import (
    Field,
    ValidationInfo,
    computed_field,
    field_validator,
    model_validator, BaseModel
)
from typing import Optional, List, Dict, Union, Literal, Tuple, Any, Sequence


class GPCCAFitModel(BaseModel):
    """
    Input schema for CellRank's GPCCA.fit method which prepares the estimator for terminal states prediction.
    """
    kernel: Literal['pseudotime', 'cytotrace', 'velocity', 'connectivity', 'realtime'] = Field(
        description="Type of kernel to use."
    )
    
    n_states: Optional[Union[int, List[int]]] = Field(
        default=None,
        description="Number of macrostates to compute. If a list, use the minChi criterion. If None, use the eigengap heuristic."
    )
    
    n_cells: Optional[int] = Field(
        default=30,
        description="Number of most likely cells from each macrostate to select."
    )
    
    cluster_key: Optional[str] = Field(
        default=None,
        description="If a key to cluster labels is given, names and colors of the states will be associated with the clusters."
    )
    
    # # Parameters for compute_schur method n_states is n_components in compute_schur
    # n_components: int = Field(
    #     default=20,
    #     description="Number of Schur vectors to compute."
    # )
    
    initial_distribution: Optional[Any] = Field(
        default=None,
        description="Input distribution over all cells. If None, uniform distribution is used."
    )
    
    method: Literal['krylov', 'brandts'] = Field(
        default='krylov',
        description="Method for calculating the Schur vectors. 'krylov' is an iterative procedure for large, sparse matrices. 'brandts' is full sorted Schur decomposition of a dense matrix."
    )
    
    which: Literal['LR', 'LM'] = Field(
        default='LR',
        description="How to sort the eigenvalues. 'LR' - the largest real part. 'LM' - the largest magnitude."
    )
    
    alpha: float = Field(
        default=1.0,
        description="Used to compute the eigengap. alpha is the weight given to the deviation of an eigenvalue from one."
    )


class GPCCAPredictInitialStatesModel(BaseModel):
    """
    Input schema for CellRank's GPCCA.predict_initial_states method which computes initial states from macrostates.
    """
    kernel: Literal['pseudotime', 'cytotrace', 'velocity', 'connectivity', 'realtime'] = Field(
        ...,
        description="Type of kernel to use."
    )
    n_states: int = Field(
        default=1,
        description="Number of initial states to compute."
    )
    
    n_cells: int = Field(
        default=30,
        description="Number of most likely cells from each macrostate to select."
    )
    
    allow_overlap: bool = Field(
        default=False,
        description="Whether to allow overlapping names between initial and terminal states."
    )


class GPCCAPredictTerminalStatesModel(BaseModel):
    """
    Input schema for CellRank's GPCCA.predict_terminal_states method which automatically selects terminal states from macrostates.
    """
    kernel: Literal['pseudotime', 'cytotrace', 'velocity', 'connectivity', 'realtime'] = Field(
        description="Type of kernel to use."
    )
    method: Literal['stability', 'top_n', 'eigengap', 'eigengap_coarse'] = Field(
        default='stability',
        description="How to select the terminal states. 'eigengap' - select based on eigengap of transition_matrix. "
                   "'eigengap_coarse' - select based on eigengap of diagonal of coarse_T. "
                   "'top_n' - select top n_states based on probability of diagonal of coarse_T. "
                   "'stability' - select states with stability >= stability_threshold."
    )
    
    n_cells: int = Field(
        default=30,
        description="Number of most likely cells from each macrostate to select."
    )
    
    alpha: float = Field(
        default=1,
        description="Weight given to the deviation of an eigenvalue from one. Only used when method = 'eigengap' or method = 'eigengap_coarse'."
    )
    
    stability_threshold: float = Field(
        default=0.96,
        description="Threshold used when method = 'stability'."
    )
    
    n_states: Optional[int] = Field(
        default=None,
        description="Number of states used when method = 'top_n'."
    )
    
    allow_overlap: bool = Field(
        default=False,
        description="Whether to allow overlapping names between initial and terminal states."
    )


class GPCCAComputeFateProbabilitiesModel(BaseModel):
    """
    Input schema for CellRank's GPCCA.compute_fate_probabilities method which calculates the probability
    of each cell being absorbed in any of the terminal states.
    """
    kernel: Literal['pseudotime', 'cytotrace', 'velocity', 'connectivity', 'realtime'] = Field(
        description="Type of kernel to use."
    )
    keys: Optional[List[str]] = Field(
        default=None,
        description="Terminal states for which to compute the fate probabilities. If None, use all states defined in terminal_states."
    )
    
    solver: Literal['direct', 'gmres', 'lgmres', 'bicgstab', 'gcrotmk'] = Field(
        default='gmres',
        description="Solver to use for the linear problem. Options are 'direct', 'gmres', 'lgmres', 'bicgstab' or 'gcrotmk' when use_petsc = False."
    )
    
    use_petsc: bool = Field(
        default=True,
        description="Whether to use solvers from petsc4py or scipy. Recommended for large problems."
    )
    
    n_jobs: Optional[int] = Field(
        default=None,
        description="Number of parallel jobs to use when using an iterative solver."
    )
    
    backend: Literal['loky', 'multiprocessing', 'threading'] = Field(
        default='loky',
        description="Which backend to use for multiprocessing."
    )
    
    tol: float = Field(
        default=1e-06,
        description="Convergence tolerance for the iterative solver."
    )
    
    preconditioner: Optional[str] = Field(
        default=None,
        description="Preconditioner to use, only available when use_petsc = True. We recommend the 'ilu' preconditioner for badly conditioned problems."
    )
    



import inspect
import os
from pathlib import Path
from scmcp_shared.util import get_env

def set_fig_path(func,  **kwargs):
    fig_dir = Path(os.getcwd()) / "figures"

    kwargs.pop("save", None)
    kwargs.pop("show", None)
    args = []
    for k,v in kwargs.items():
        if isinstance(v, (tuple, list, set)):
            args.append(f"{k}-{'-'.join([str(i) for i in v])}")
        else:
            args.append(f"{k}-{v}")
    args_str = "_".join(args)
    if func == "scvelo_projection":
        old_path = fig_dir / f"scvelo_{kwargs['kernel']}.png"
        fig_path = fig_dir / f"{func}_{args_str}.png"
    else:
        if (fig_dir / f"{func}_.png").is_file():
            old_path = fig_dir / f"{func}_.png"
        else:
            old_path = fig_dir / f"{func}.png"
        fig_path = fig_dir / f"{func}_{args_str}.png"
    try:
        os.rename(old_path, fig_path)
        return fig_path
    except FileNotFoundError:
        print(f"The file {old_path} does not exist")
    except FileExistsError:
        print(f"The file {fig_path} already exists")
    except PermissionError:
        print("You don't have permission to rename this file")
    transport = get_env("TRANSPORT") 
    if transport == "stdio":
        return fig_path
    else:
        host = get_env("HOST")
        port = get_env("PORT")
        fig_path = f"http://{host}:{port}/figures/{Path(fig_path).name}"
        return fig_path


import scvelo as scv
import scanpy as sc
import inspect
from ...schema.pp import *
from scmcp_shared.util import filter_args, add_op_log, forward_request, get_ads
from scmcp_shared.logging_config import setup_logger
from fastmcp.exceptions import ToolError
from scmcp_shared.schema import AdataInfo
from scmcp_shared.server.preset import ScanpyPreprocessingMCP


logger = setup_logger()


pp_mcp = ScanpyPreprocessingMCP(include_tools=["neighbors"]).mcp


@pp_mcp.tool()
def filter_and_normalize(
    request: FilterAndNormalizeModel, adinfo: AdataInfo = AdataInfo()
):
    """
    Preprocess data: filter and normalize AnnData object for velocity/pseudotime analysis.
    """
    try:
        result = forward_request("filter_and_normalize", request, adinfo)
        if result is not None:
            return result
        ads = get_ads()
        adata = ads.get_adata(adinfo=adinfo).copy()
        if "spliced" in adata.layers:
            # Filter arguments based on the function parameters
            kwargs = filter_args(request, scv.pp.filter_and_normalize)
            if "log" in inspect.signature(scv.pp.filter_and_normalize).parameters:
                kwargs["log"] = False
            # Run filter_and_normalize
            scv.pp.filter_and_normalize(adata, **kwargs)
            sc.pp.log1p(adata)
            add_op_log(adata, scv.pp.filter_and_normalize, kwargs, adinfo)
            add_op_log(adata, sc.pp.log1p, {}, adinfo)
        else:
            sc.pp.filter_genes(adata, **filter_args(request, sc.pp.filter_genes))
            sc.pp.normalize_total(adata, **filter_args(request, sc.pp.normalize_total))
            sc.pp.log1p(adata)
            sc.pp.highly_variable_genes(
                adata, **filter_args(request, sc.pp.highly_variable_genes)
            )
            adata.layers["spliced"] = adata.X
            adata.layers["unspliced"] = adata.X
        ads.set_adata(adata, adinfo=adinfo)
        return {
            "sampleid": adinfo.sampleid or ads.active_id,
            "dtype": adinfo.adtype,
            "adata": adata,
        }

    except ToolError as e:
        raise ToolError(e)
    except Exception as e:
        if hasattr(e, "__") and e.__:
            raise ToolError(e.__context__)
        else:
            raise ToolError(e)



"""
Command-line interface for cellrank-mcp.

This module provides a CLI entry point for the cellrank-mcp package.
"""

from scmcp_shared.cli import MCPCLI
from .server import CellrankMCPManager

cli = MCPCLI(
    name="cellrank-mcp", 
    help_text="Cellrank MCP Server CLI",
    manager=CellrankMCPManager
)



import pytest


@pytest.fixture
def mcp():
    from cellrank_mcp.server import CellrankMCPManager
    from scmcp_shared.backend import AdataManager

    mcp = CellrankMCPManager("cellrank-mcp", backend=AdataManager).mcp
    return mcp



from fastmcp.exceptions import ToolError
from ...schema.pl import *
import cellrank as cr
from scmcp_shared.util import filter_args, forward_request, get_ads
from scmcp_shared.logging_config import setup_logger
from ...util import set_fig_path
from scmcp_shared.schema import AdataInfo
from scmcp_shared.server.preset import ScanpyPlottingMCP

logger = setup_logger()


pl_mcp = ScanpyPlottingMCP(
    exclude_tools=[
        "tracksplot",
        "rank_genes_groups",
        "rank_genes_groups_dotplot",
        "violin",
        "heatmap",
        "dotplot",
        "matrixplot",
        "highly_variable_genes",
    ],
).mcp


@pl_mcp.tool()
def kernel_projection(
    request: KernelPlotProjectionModel, adinfo: AdataInfo = AdataInfo()
):
    """Plot transition matrix as a stream or grid plot for a specified kernel."""
    try:
        result = forward_request("kernel_projection", request, adinfo)
        if result is not None:
            return result
        kernel_type = request.kernel
        ads = get_ads()
        kernel = ads.cr_kernel[kernel_type]

        # Check if transition matrix has been computed
        if not hasattr(kernel, "transition_matrix") or kernel.transition_matrix is None:
            error_msg = f"Transition matrix for kernel '{kernel_type}' has not been computed. Please compute transition matrix first."
            logger.error(error_msg)
            raise ValueError(error_msg)

        # Filter arguments based on the plot_projection method
        kwargs = filter_args(request, kernel.plot_projection)
        kwargs["save"] = kernel_type
        kwargs["show"] = False
        kwargs["color"] = request.color
        kwargs["legend_loc"] = request.legend_loc
        # Plot the projection
        kernel.plot_projection(**kwargs)
        kwargs["kernel"] = kernel_type
        fig_path = set_fig_path("scvelo_projection", **kwargs)
        return {"figpath": fig_path}

    except ToolError as e:
        raise ToolError(e)
    except Exception as e:
        if hasattr(e, "__context__") and e.__context__:
            raise ToolError(e.__context__)
        else:
            raise ToolError(e)


@pl_mcp.tool()
def circular_projection(
    request: CircularProjectionModel, adinfo: AdataInfo = AdataInfo()
):
    """
    Visualize fate probabilities in a circular embedding. compute_fate_probabilities first.
    """
    try:
        result = forward_request("circular_projection", request, adinfo)
        if result is not None:
            return result
        # Check if AnnData object exists in the session
        ads = get_ads()
        adata = ads.get_adata(adinfo=adinfo)

        if "lineages_fwd" not in adata.obsm:
            raise ValueError(
                "No lineages_fwd found. Please call compute_fate_probabilities first."
            )

        # Filter arguments based on the circular_projection function
        kwargs = filter_args(request, cr.pl.circular_projection)

        kwargs["save"] = "circular_projection.png"
        kwargs["legend_loc"] = "right"
        kwargs["keys"] = request.keys
        # Call the circular_projection function
        cr.pl.circular_projection(adata, **kwargs)
        del kwargs["save"]
        fig_path = set_fig_path("circular_projection", **kwargs)
        return {
            "status": "success",
            "message": "Successfully created circular projection plot",
            "figpath": fig_path,
        }
    except ToolError as e:
        raise ToolError(e)
    except Exception as e:
        if hasattr(e, "__context__") and e.__context__:
            raise ToolError(e.__context__)
        else:
            raise ToolError(e)



from pydantic import (
    Field,
    ValidationInfo,
    computed_field,
    field_validator,
    model_validator,
    BaseModel
)
from typing import Optional, List, Dict, Union, Literal, Tuple, Any


class FilterAndNormalizeModel(BaseModel):
    """Input schema for filter_and_normalize preprocessing function."""
    
    min_counts: Optional[int] = Field(
        default=None,
        description="Minimum number of counts required for a gene to pass filtering (spliced)."
    )
    min_counts_u: Optional[int] = Field(
        default=None,
        description="Minimum number of counts required for a gene to pass filtering (unspliced)."
    )
    min_cells: Optional[int] = Field(
        default=None,
        description="Minimum number of cells expressed required to pass filtering (spliced)."
    )
    min_cells_u: Optional[int] = Field(
        default=None,
        description="Minimum number of cells expressed required to pass filtering (unspliced)."
    )
    min_shared_counts: Optional[int] = Field(
        default=None,
        description="Minimum number of counts (both unspliced and spliced) required for a gene."
    )
    min_shared_cells: Optional[int] = Field(
        default=None,
        description="Minimum number of cells required to be expressed (both unspliced and spliced)."
    )
    n_top_genes: Optional[int] = Field(
        default=None,
        description="Number of genes to keep."
    )
    retain_genes: Optional[List[str]] = Field(
        default=None,
        description="List of gene names to be retained independent of thresholds."
    )
    subset_highly_variable: bool = Field(
        default=True,
        description="Whether to subset highly variable genes or to store in .var['highly_variable']."
    )
    
    flavor: Literal["seurat", "cell_ranger", "svr"] = Field(
        default="seurat",
        description="Choose the flavor for computing normalized dispersion. If choosing 'seurat', this expects non-logarithmized data."
    )
    
    layers_normalize: Optional[List[str]] = Field(
        default=None,
        description="List of layers to be normalized. If set to None, the layers {'X', 'spliced', 'unspliced'} are considered for normalization."
    )
    target_sum: Optional[float] = Field(
            default=None,
            description="If None, after normalization, each observation (cell) has a total count equal to the median of total counts for observations (cells) before normalization."
        )



from scmcp_shared.server.preset import (
    io_mcp,
    # ScanpyPreprocessingMCP,
    ScanpyToolsMCP,
    # ScanpyPlottingMCP,
    ScanpyUtilMCP,
)
from scmcp_shared.mcp_base import BaseMCPManager

from .preset.pl import pl_mcp
from .preset.kernel import kernel_mcp
from .preset.pp import pp_mcp
from .preset.estimator import estimator_mcp

# from .code.rag import rag_mcp
from scmcp_shared.server.code import nb_mcp

tl_mcp = ScanpyToolsMCP().mcp
ul_mcp = ScanpyUtilMCP(
    include_tools=["query_op_log", "check_samples"],
).mcp


class CellrankMCPManager(BaseMCPManager):
    """Manager class for Cellrank MCP modules."""

    def init_mcp(self):
        """Initialize available Cellrank MCP modules."""
        self.available_modules = {
            "io": io_mcp,
            "pp": pp_mcp,
            "kernel": kernel_mcp,
            "estimator": estimator_mcp,
            "pl": pl_mcp,
            "tl": tl_mcp,
            "ul": ul_mcp,
            # "rag": rag_mcp,
            "nb": nb_mcp,
        }



from pydantic import (
    Field,
    ValidationInfo,
    computed_field,
    field_validator,
    model_validator,BaseModel
)
from typing import Optional, List, Dict, Union, Literal, Any, Tuple


class PseudotimeKernelModel(BaseModel):
    """Input schema for CellRank's PseudotimeKernel which computes directed transition probabilities based on a k-NN graph and pseudotime."""
    
    time_key: str = Field(
        description="Key in AnnData.obs where the pseudotime is stored."
    )
    backward: bool = Field(
        default=False,
        description="Direction of the process. If True, the pseudotime will be set to max(pseudotime) - pseudotime."
    )


class CytoTRACEKernelModel(BaseModel):
    """Input schema for CellRank's CytoTRACEKernel which computes directed transition probabilities using the CytoTRACE score."""
    
    backward: bool = Field(
        default=False,
        description="Direction of the process. If True, the CytoTRACE score will be inverted."
    )


class VelocityKernelModel(BaseModel):
    """Input schema for CellRank's VelocityKernel which computes a transition matrix based on RNA velocity."""
    
    backward: bool = Field(
        default=False,
        description="Direction of the process. If True, velocity vectors will be reversed."
    )
    
    attr: Optional[Literal['layers', 'obsm']] = Field(
        default='layers',
        description="Attribute of AnnData to read from."
    )
    
    xkey: Optional[str] = Field(
        default='Ms',
        description="Key in AnnData.layers or AnnData.obsm where expected gene expression counts are stored."
    )
    
    vkey: Optional[str] = Field(
        default='velocity',
        description="Key in AnnData.layers or AnnData.obsm where velocities are stored."
    )
    
    gene_subset: Optional[List[str]] = Field(
        default=None,
        description="List of genes to be used to compute transition probabilities. If not specified, genes from adata.var['{vkey}_genes'] are used."
    )


class ConnectivityKernelModel(BaseModel):
    """Input schema for CellRank's ConnectivityKernel which computes transition probabilities based on similarities among cells."""
    
    conn_key: str = Field(
        default='connectivities',
        description="Key in AnnData.obsp where connectivity matrix describing cell-cell similarity is stored."
    )
    
    check_connectivity: bool = Field(
        default=False,
        description="Check whether the underlying kNN graph is connected."
    )


class RealTimeKernelModel(BaseModel):
    """Input schema for creating a RealTimeKernel"""
    
    time_key: str = Field(
        description="Key in AnnData.obs containing the experimental time."
    )
    # Add couplings field
    couplings: Optional[Dict[Tuple[Any, Any], Any]] = Field(
        default=None,
        description="Pre-computed transport couplings. The keys should correspond to a tuple of categories from the time. If None, the keys will be constructed using the policy and the compute_coupling() method must be overriden."
    )
    policy: Literal['sequential', 'triu'] = Field(
        default='sequential',
        description="How to construct keys from time: 'sequential' for [(t1,t2),(t2,t3),...], 'triu' for [(t1,t2),(t1,t3),...,(t2,t3),...]."
    )
    


class KernelModel(BaseModel):
    """Unified input schema for creating various CellRank Kernels."""
    
    kernel: Literal['pseudotime', 'cytotrace', 'velocity', 'connectivity', 'realtime'] = Field(
        description="Type of kernel to use."
    )
    
    # PseudotimeKernel and RealTimeKernel specific fields
    time_key: Optional[str] = Field(
        default=None,
        description="[pseudotime, realtime] Key in AnnData.obs where the pseudotime or experimental time is stored."
    )
    
    # Shared fields across multiple kernels
    backward: Optional[bool] = Field(
        default=False,
        description="[pseudotime, cytotrace, velocity] Direction of the process. If True, the direction will be reversed."
    )
    
    # VelocityKernel specific fields
    attr: Optional[Literal['layers', 'obsm']] = Field(
        default='layers',
        description="[velocity] Attribute of AnnData to read from."
    )
    
    xkey: Optional[str] = Field(
        default='Ms',
        description="[velocity] Key in AnnData.layers or AnnData.obsm where expected gene expression counts are stored."
    )
    
    vkey: Optional[str] = Field(
        default='velocity',
        description="[velocity] Key in AnnData.layers or AnnData.obsm where velocities are stored."
    )
    
    gene_subset: Optional[List[str]] = Field(
        default=None,
        description="[velocity] List of genes to be used to compute transition probabilities. If not specified, genes from adata.var['{vkey}_genes'] are used."
    )
    
    # ConnectivityKernel specific fields
    conn_key: Optional[str] = Field(
        default='connectivities',
        description="[connectivity] Key in AnnData.obsp where connectivity matrix describing cell-cell similarity is stored."
    )
    
    check_connectivity: Optional[bool] = Field(
        default=False,
        description="[connectivity] Check whether the underlying kNN graph is connected."
    )
    
    # RealTimeKernel specific fields
    couplings: Optional[Dict[Tuple[Any, Any], Any]] = Field(
        default=None,
        description="[realtime] Pre-computed transport couplings. The keys should correspond to a tuple of categories from the time. If None, the keys will be constructed using the policy and the compute_coupling() method must be overriden."
    )
    
    policy: Optional[Literal['sequential', 'triu']] = Field(
        default='sequential',
        description="[realtime] How to construct keys from time: 'sequential' for [(t1,t2),(t2,t3),...], 'triu' for [(t1,t2),(t1,t3),...,(t2,t3),...]."
    )



class VelocityComputeTransitionMatrixModel(BaseModel):
    """Input schema for computing transition matrix based on velocity directions on the local manifold in CellRank."""
    
    model: Literal['deterministic', 'stochastic', 'monte_carlo'] = Field(
        default='deterministic',
        description="How to compute transition probabilities. 'deterministic' doesn't propagate uncertainty, 'monte_carlo' uses random sampling, 'stochastic' uses second order approximation (requires jax)."
    )
    
    backward_mode: Literal['transpose', 'negate'] = Field(
        default='transpose',
        description="Only matters if kernel initialized with backward=True. 'transpose' computes transitions from neighboring cells to cell i, 'negate' negates the velocity vector."
    )
    
    similarity: Literal['correlation', 'cosine', 'dot_product'] = Field(
        default='correlation',
        description="Similarity measure between cells as described in Li et al., 2021."
    )
    
    softmax_scale: Optional[float] = Field(
        default=None,
        description="Scaling parameter for the softmax. If None, estimated using 1 / median(correlations) to counter high-dimensional orthogonality."
    )
    
    n_samples: int = Field(
        default=1000,
        description="Number of samples when model='monte_carlo'."
    )
    
    seed: Optional[int] = Field(
        default=None,
        description="Random seed when model='monte_carlo'."
    )
    
    n_jobs: Optional[int] = Field(
        default=None,
        description="Number of parallel jobs. If -1, use all available cores. If None or 1, execution is sequential."
    )
    
    backend: Literal['loky', 'multiprocessing', 'threading'] = Field(
        default='loky',
        description="Which backend to use for parallelization."
    )


class ConnectivityComputeTransitionMatrixModel(BaseModel):
    """Input schema for computing transition matrix based on transcriptomic similarity in CellRank's ConnectivityKernel."""
    
    density_normalize: bool = Field(
        default=True,
        description="Whether to use the underlying kNN graph for density normalization."
    )


class PseudotimeComputeTransitionMatrixModel(BaseModel):
    """Input schema for computing transition matrix based on k-NN graph and pseudotemporal ordering in CellRank's PseudotimeKernel."""
    
    threshold_scheme: Literal['soft', 'hard'] = Field(
        default='hard',
        description="Which method to use when biasing the graph. 'hard' removes edges against pseudotime direction (Palantir), 'soft' down-weights them (VIA)."
    )
    
    frac_to_keep: float = Field(
        default=0.3,
        description="Fraction of closest neighbors to keep regardless of pseudotime. Only used with 'hard' scheme. Must be in [0, 1]."
    )
    
    b: float = Field(
        default=10.0,
        description="Growth rate of generalized logistic function. Only used with 'soft' scheme."
    )
    
    nu: float = Field(
        default=0.5,
        description="Affects near which asymptote maximum growth occurs. Only used with 'soft' scheme."
    )
    
    check_irreducibility: bool = Field(
        default=False,
        description="Whether to check for irreducibility of the final transition matrix."
    )
    
    n_jobs: Optional[int] = Field(
        default=None,
        description="Number of parallel jobs. If -1, use all available cores. If None or 1, execution is sequential."
    )
    
    backend: Literal['loky', 'multiprocessing', 'threading'] = Field(
        default='loky',
        description="Which backend to use for parallelization."
    )
    

class CytoTRACEComputeTransitionMatrixModel(BaseModel):
    """Input schema for computing transition matrix based on k-NN graph and CytoTRACE scores in CellRank's CytoTRACEKernel."""
    
    threshold_scheme: Literal['soft', 'hard'] = Field(
        default='hard',
        description="Which method to use when biasing the graph. 'hard' removes edges against CytoTRACE direction (Palantir), 'soft' down-weights them (VIA)."
    )
    
    frac_to_keep: float = Field(
        default=0.3,
        description="Fraction of closest neighbors to keep regardless of CytoTRACE score. Only used with 'hard' scheme. Must be in [0, 1]."
    )
    
    b: float = Field(
        default=10.0,
        description="Growth rate of generalized logistic function. Only used with 'soft' scheme."
    )
    
    nu: float = Field(
        default=0.5,
        description="Affects near which asymptote maximum growth occurs. Only used with 'soft' scheme."
    )
    
    check_irreducibility: bool = Field(
        default=False,
        description="Whether to check for irreducibility of the final transition matrix."
    )
    
    n_jobs: Optional[int] = Field(
        default=None,
        description="Number of parallel jobs. If -1, use all available cores. If None or 1, execution is sequential."
    )
    
    backend: Literal['loky', 'multiprocessing', 'threading'] = Field(
        default='loky',
        description="Which backend to use for parallelization."
    )
    

class RealTimeComputeTransitionMatrixModel(BaseModel):
    """Input schema for computing transition matrix from optimal transport couplings in CellRank's RealTimeKernel."""
    
    threshold: Union[int, float, Literal['auto', 'auto_local'], None] = Field(
        default='auto',
        description="How to remove small non-zero values from the transition matrix. 'auto' finds maximum threshold that keeps at least one non-zero per row, 'auto_local' does this per transport, float is percentage of non-zeros to remove."
    )
    
    self_transitions: Union[Literal['uniform', 'diagonal', 'connectivities', 'all'], List[Any]] = Field(
        default='connectivities',
        description="How to define transitions within blocks corresponding to same time point. 'uniform' for row-normalized matrix, 'diagonal' for identity, 'connectivities' for ConnectivityKernel transitions, 'all' or sequence for specific blocks."
    )
    
    conn_weight: Optional[float] = Field(
        default=None,
        description="Weight of connectivities' self transitions. Only used when self_transitions='all' or a sequence of source keys is passed."
    )
    
    conn_kwargs: Dict[str, Any] = Field(
        default={},
        description="Keyword arguments for neighbors() or compute_transition_matrix() when using self_transitions='connectivities'."
    )
    

class ComputeTransitionMatrixModel(BaseModel):
    """Unified input schema for computing transition matrices across different CellRank kernel types."""
    kernel: Literal['pseudotime', 'cytotrace', 'velocity', 'connectivity', 'realtime'] = Field(
        description="Type of kernel to use."
    )
    
    # Common parameters across multiple kernels
    n_jobs: Optional[int] = Field(
        default=None,
        description="[velocity, pseudotime, cytotrace] Number of parallel jobs. If -1, use all available cores. If None or 1, execution is sequential."
    )
    
    backend: Literal['loky', 'multiprocessing', 'threading'] = Field(
        default='loky',
        description="[velocity, pseudotime, cytotrace] Which backend to use for parallelization."
    )
    
    check_irreducibility: bool = Field(
        default=False,
        description="[pseudotime, cytotrace] Whether to check for irreducibility of the final transition matrix."
    )
    
    # VelocityKernel specific parameters
    model: Optional[Literal['deterministic', 'stochastic', 'monte_carlo']] = Field(
        default=None,
        description="[velocity] How to compute transition probabilities. 'deterministic' doesn't propagate uncertainty, 'monte_carlo' uses random sampling, 'stochastic' uses second order approximation (requires jax)."
    )
    
    backward_mode: Optional[Literal['transpose', 'negate']] = Field(
        default=None,
        description="[velocity] Only matters if kernel initialized with backward=True. 'transpose' computes transitions from neighboring cells to cell i, 'negate' negates the velocity vector."
    )
    
    similarity: Optional[Literal['correlation', 'cosine', 'dot_product']] = Field(
        default=None,
        description="[velocity] Similarity measure between cells as described in Li et al., 2021."
    )
    
    softmax_scale: Optional[float] = Field(
        default=None,
        description="[velocity] Scaling parameter for the softmax. If None, estimated using 1 / median(correlations) to counter high-dimensional orthogonality."
    )
    
    n_samples: Optional[int] = Field(
        default=None,
        description="[velocity] Number of samples when model='monte_carlo'."
    )
    
    # ConnectivityKernel specific parameters
    density_normalize: Optional[bool] = Field(
        default=None,
        description="[connectivity] Whether to use the underlying kNN graph for density normalization."
    )
    
    # PseudotimeKernel and CytoTRACEKernel shared parameters
    threshold_scheme: Optional[Literal['soft', 'hard']] = Field(
        default=None,
        description="[pseudotime, cytotrace] Which method to use when biasing the graph. 'hard' removes edges against direction, 'soft' down-weights them."
    )
    
    frac_to_keep: Optional[float] = Field(
        default=None,
        description="[pseudotime, cytotrace] Fraction of closest neighbors to keep regardless of ordering. Only used with 'hard' scheme. Must be in [0, 1]."
    )
    
    b: Optional[float] = Field(
        default=None,
        description="[pseudotime, cytotrace] Growth rate of generalized logistic function. Only used with 'soft' scheme."
    )
    
    nu: Optional[float] = Field(
        default=None,
        description="[pseudotime, cytotrace] Affects near which asymptote maximum growth occurs. Only used with 'soft' scheme."
    )
    
    # RealTimeKernel specific parameters
    threshold: Optional[Union[int, float, Literal['auto', 'auto_local'], None]] = Field(
        default=None,
        description="[realtime] How to remove small non-zero values from the transition matrix. 'auto' finds maximum threshold that keeps at least one non-zero per row."
    )
    
    self_transitions: Optional[Union[Literal['uniform', 'diagonal', 'connectivities', 'all'], List[Any]]] = Field(
        default=None,
        description="[realtime] How to define transitions within blocks corresponding to same time point."
    )
    
    conn_weight: Optional[float] = Field(
        default=None,
        description="[realtime] Weight of connectivities' self transitions. Only used when self_transitions='all' or a sequence of source keys is passed."
    )
    
    conn_kwargs: Optional[Dict[str, Any]] = Field(
        default=None,
        description="[realtime] Keyword arguments for neighbors() or compute_transition_matrix() when using self_transitions='connectivities'."
    )



from fastmcp import FastMCP
from fastmcp.exceptions import ToolError
import cellrank as cr
import scvelo as scv
from ...schema.kernel import *
from scmcp_shared.util import filter_args, forward_request, get_ads
from scmcp_shared.logging_config import setup_logger
from scmcp_shared.schema import AdataInfo

logger = setup_logger()


kernel_mcp = FastMCP("CellrankMCP-Kernel-Server")


@kernel_mcp.tool()
def create_kernel(request: KernelModel, adinfo: AdataInfo = AdataInfo()):
    """Create a CellRank kernel based on the specified type and parameters."""
    try:
        # Check if AnnData object exists in the session
        result = forward_request("kernel_create_kernel", request, adinfo)
        if result is not None:
            return result
        ads = get_ads()
        adata = ads.get_adata(adinfo=adinfo).copy()
        kernel_type = request.kernel

        kernel_dic = {
            "pseudotime": cr.kernels.PseudotimeKernel,
            "cytotrace": cr.kernels.CytoTRACEKernel,
            "velocity": cr.kernels.VelocityKernel,
            "connectivity": cr.kernels.ConnectivityKernel,
            "realtime": cr.kernels.RealTimeKernel,
        }
        if kernel_type not in kernel_dic:
            return {
                "status": "error",
                "message": f"Unsupported kernel type: {kernel_type}",
            }
        kernel_class = kernel_dic[kernel_type]
        kwargs = filter_args(request, kernel_class)
        if kernel_type == "cytotrace":
            if "Ms" not in adata.layers or "Mu" not in adata.layers:
                scv.pp.moments(adata, n_pcs=30, n_neighbors=30)
            kernel = kernel_class(adata=adata, **kwargs).compute_cytotrace()
        else:
            kernel = kernel_class(adata=adata, **kwargs)

        ads.cr_kernel[kernel_type] = kernel
        ads.set_adata(adata, adinfo=adinfo)
        return {
            "status": "success",
            "message": f"Successfully created {kernel_type} kernel",
            "kernel_type": kernel_type,
        }

    except ToolError as e:
        raise ToolError(e)
    except Exception as e:
        if hasattr(e, "__context__") and e.__context__:
            raise ToolError(e.__context__)
        else:
            raise ToolError(e)


@kernel_mcp.tool()
def compute_transition_matrix(
    request: ComputeTransitionMatrixModel, adinfo: AdataInfo = AdataInfo()
):
    """Compute transition matrix for a specified kernel using appropriate parameters."""
    try:
        result = forward_request("compute_transition_matrix", request, adinfo)
        if result is not None:
            return result
        kernel_type = request.kernel
        ads = get_ads()
        kernel = ads.cr_kernel[kernel_type].copy()
        kwargs = filter_args(
            request, kernel.compute_transition_matrix, show_progress_bar=False
        )
        kernel.compute_transition_matrix(**kwargs)
        ads.cr_kernel[kernel_type] = kernel
        return {
            "status": "success",
            "message": f"Successfully computed transition matrix for {kernel_type} kernel",
        }

    except ToolError as e:
        raise ToolError(e)
    except Exception as e:
        if hasattr(e, "__context__") and e.__context__:
            raise ToolError(e.__context__)
        else:
            raise ToolError(e)



from pydantic import (
    Field,
    ValidationInfo,
    computed_field,
    field_validator,
    model_validator,
    BaseModel
)
from typing import Optional, List, Dict, Union, Literal, Tuple, Any



class KernelPlotProjectionModel(BaseModel):
    """Input schema for plotting transition matrix as a stream or grid plot in CellRank kernels."""
    
    basis: str = Field(
        default='umap',
        description="Key in obsm containing the basis."
    )
    
    key_added: Optional[str] = Field(
        default=None,
        description="If not None, save the result to adata.obsm['{key_added}']. Otherwise, save the result to 'T_fwd_{basis}' or 'T_bwd_{basis}', depending on the direction."
    )
    
    recompute: bool = Field(
        default=False,
        description="Whether to recompute the projection if it already exists."
    )
    
    stream: bool = Field(
        default=True,
        description="If True, use velocity_embedding_stream(). Otherwise, use velocity_embedding_grid()."
    )
    
    connectivities: Optional[Any] = Field(
        default=None,
        description="Connectivity matrix to use for projection. If None, use ones from the underlying kernel, if possible."
    )
    
    color: Optional[Union[str, List[str]]] = Field(
        default=None,
        description="Key for annotations of observations/cells or variables/genes."
    )
    
    legend_loc: str = Field(
        default=None,
        description="Location of legend, either 'on data', 'right margin' or valid keywords for matplotlib.legend."
    )
    
    kernel: Literal['pseudotime', 'cytotrace', 'velocity', 'connectivity', 'realtime'] = Field(
        description="Type of kernel to use."
    )


class CircularProjectionModel(BaseModel):
    """
    Input schema for CellRank's circular_projection function which visualizes fate probabilities in a circular embedding.
    """
    
    keys: Union[str, List[str]] = Field(
        description="Keys in obs or var_names. Can include 'kl_divergence' or 'entropy'."
    )
    
    backward: bool = Field(
        default=False,
        description="Direction of the process."
    )
    
    lineages: Optional[Union[str, List[str]]] = Field(
        default=None,
        description="Lineages to plot. If None, plot all lineages."
    )
    
    early_cells: Optional[Union[Dict[str, List[str]], List[str]]] = Field(
        default=None,
        description="Cell ids or a mask marking early cells used to define the average fate probabilities. If None, use all cells."
    )
    
    lineage_order: Optional[Literal['default', 'optimal']] = Field(
        default=None,
        description="How to order lineages. 'optimal' solves the Traveling salesman problem. 'default' uses the order specified by lineages."
    )
    
    metric: Union[str, Any] = Field(
        default='correlation',
        description="Metric to use when constructing pairwise distance matrix when lineage_order = 'optimal'."
    )
    
    normalize_by_mean: bool = Field(
        default=True,
        description="If True, normalize each lineage by its mean probability."
    )
    
    ncols: int = Field(
        default=4,
        description="Number of columns when plotting multiple keys."
    )
    
    space: float = Field(
        default=0.25,
        description="Horizontal and vertical space between for subplots_adjust()."
    )
    
    use_raw: bool = Field(
        default=False,
        description="Whether to access raw when there are keys in var_names."
    )
    
    text_kwargs: Dict[str, Any] = Field(
        default={},
        description="Keyword arguments for text()."
    )
    
    label_distance: float = Field(
        default=1.25,
        description="Distance at which the lineage labels will be drawn."
    )
    
    label_rot: Union[Literal['default', 'best'], float] = Field(
        default='best',
        description="How to rotate the labels. 'best' rotates for readability, 'default' uses matplotlib's default, float rotates by degrees."
    )
    
    show_edges: bool = Field(
        default=True,
        description="Whether to show the edges surrounding the simplex."
    )
    
    key_added: Optional[str] = Field(
        default=None,
        description="Key in obsm where to add the circular embedding. If None, it will be set to 'X_fate_simplex_{fwd,bwd}'."
    )
    
    figsize: Optional[Tuple[float, float]] = Field(
        default=None,
        description="Size of the figure."
    )
    
    dpi: Optional[int] = Field(
        default=None,
        description="Dots per inch."
    )
    
    save: Optional[Union[str, bool]] = Field(
        default=None,
        description="Filename where to save the plot."
    )
    
    # Common plotting parameters
    color: Optional[Union[str, List[str]]] = Field(
        default=None,
        description="Key for annotations of observations/cells or variables/genes."
    )
    
    alpha: Optional[float] = Field(
        default=None,
        description="Alpha value for the scatter plot."
    )
    
    size: Optional[Union[float, int]] = Field(
        default=None,
        description="Point size for the scatter plot."
    )
    
    legend_loc: Optional[str] = Field(
        default="right",
        description="Location of legend, either 'on data', 'right margin' or valid keywords for matplotlib.legend."
    )
    
    title: Optional[Union[str, List[str]]] = Field(
        default=None,
        description="Title for the plot."
    )


from fastmcp import FastMCP
from fastmcp.exceptions import ToolError
import cellrank as cr
from ...schema.estimator import *
from scmcp_shared.util import filter_args, forward_request, get_ads
from scmcp_shared.logging_config import setup_logger
from scmcp_shared.schema import AdataInfo

logger = setup_logger()

estimator_mcp = FastMCP("CellrankMCP-Estimator-Server")


@estimator_mcp.tool()
def create_and_fit_gpcca(request: GPCCAFitModel, adinfo: AdataInfo = AdataInfo()):
    """Generalized Perron Cluster Cluster Analysis (GPCCA), Use it to compute macrostates.
    Need to compute transition matrix first as `.compute_transition_matrix()
    """
    try:
        result = forward_request("create_and_fit_gpcca", request, adinfo)
        if result is not None:
            return result
        kernel_type = request.kernel

        # ads = get_ads()
        # adata = ads.get_adata(adinfo=adinfo)

        kernel = ads.cr_kernel[kernel_type]
        estimator = cr.estimators.GPCCA(kernel)
        kwargs = filter_args(request, estimator.fit)
        schur_kwargs = filter_args(request, cr.estimators.GPCCA.compute_schur)
        kwargs.update(schur_kwargs)
        estimator.fit(**kwargs)
        ads.cr_estimator[kernel_type] = estimator
        return {
            "status": "success",
            "message": f"Successfully created and fitted the GPCCA estimator for kernel '{kernel_type}'",
            "estimator": estimator,
        }
    except ToolError as e:
        raise ToolError(e)
    except Exception as e:
        if hasattr(e, "__context__") and e.__context__:
            raise ToolError(e.__context__)
        else:
            raise ToolError(e)


@estimator_mcp.tool()
def predict_terminal_states(
    request: GPCCAPredictTerminalStatesModel, adinfo: AdataInfo = AdataInfo()
):
    """
    Predict terminal states from macrostates. This automatically selects terminal states from the computed macrostates.
    """
    try:
        result = forward_request("predict_terminal_states", request, adinfo)
        if result is not None:
            return result
        kernel_type = request.kernel
        ads = get_ads()
        estimator = ads.cr_estimator[kernel_type].copy()

        # Filter arguments based on the predict_terminal_states method
        kwargs = filter_args(request, estimator.predict_terminal_states)

        # Predict terminal states
        estimator.predict_terminal_states(**kwargs)

        # Get information about the terminal states
        terminal_states = (
            estimator.terminal_states.cat.categories
            if hasattr(estimator, "terminal_states")
            else []
        )
        ads.cr_estimator[kernel_type] = estimator
        return {
            "status": "success",
            "estimator": estimator,
            "kernel_type": kernel_type,
            "terminal_states": terminal_states,
        }
    except ToolError as e:
        raise ToolError(e)
    except Exception as e:
        if hasattr(e, "__context__") and e.__context__:
            raise ToolError(e.__context__)
        else:
            raise ToolError(e)


@estimator_mcp.tool()
def predict_initial_states(
    request: GPCCAPredictInitialStatesModel, adinfo: AdataInfo = AdataInfo()
):
    """
    Compute initial states from macrostates using coarse_stationary_distribution.
    """
    try:
        result = forward_request("predict_initial_states", request, adinfo)
        if result is not None:
            return result
        ads = get_ads()
        estimator = ads.cr_estimator[kernel_type].copy()
        # Filter arguments based on the predict_initial_states method
        kwargs = filter_args(request, estimator.predict_initial_states)

        # Predict initial states
        estimator.predict_initial_states(**kwargs)

        # Get information about the initial states
        initial_states = (
            estimator.initial_states.cat.categories
            if hasattr(estimator, "initial_states")
            else []
        )
        ads.cr_estimator[kernel_type] = estimator
        return {
            "status": "success",
            "message": f"Successfully predicted initial states for kernel '{kernel_type}'",
            "kernel_type": kernel_type,
            "initial_states": initial_states,
        }
    except ToolError as e:
        raise ToolError(e)
    except Exception as e:
        if hasattr(e, "__context__") and e.__context__:
            raise ToolError(e.__context__)
        else:
            raise ToolError(e)


@estimator_mcp.tool()
def compute_fate_probabilities(
    request: GPCCAComputeFateProbabilitiesModel, adinfo: AdataInfo = AdataInfo()
):
    """Compute fate probabilities for cells."""
    try:
        result = forward_request("compute_fate_probabilities", request, adinfo)
        if result is not None:
            return result
        kernel_type = request.kernel
        ads = get_ads()
        estimator = ads.cr_estimator[kernel_type].copy()

        # Check if terminal states have been computed
        if (
            not hasattr(estimator, "terminal_states")
            or estimator.terminal_states is None
        ):
            raise ValueError(
                "Terminal states have not been computed. Please run predict_terminal_states first."
            )

        # Filter arguments based on the compute_fate_probabilities method
        kwargs = filter_args(request, estimator.compute_fate_probabilities)
        kwargs["show_progress_bar"] = False
        # Compute fate probabilities
        estimator.compute_fate_probabilities(**kwargs)

        # Get information about the fate probabilities
        fate_probs_computed = hasattr(estimator, "fate_probabilities")
        terminal_states = (
            estimator.terminal_states.cat.categories.tolist()
            if hasattr(estimator, "terminal_states")
            else []
        )
        ads.cr_estimator[kernel_type] = estimator
        return {
            "status": "success",
            "message": f"Successfully computed fate probabilities for kernel '{kernel_type}'",
            "kernel_type": kernel_type,
            "terminal_states": terminal_states,
            "fate_probabilities_computed": fate_probs_computed,
            "estimator": estimator,
        }
    except ToolError as e:
        raise ToolError(e)
    except Exception as e:
        if hasattr(e, "__context__") and e.__context__:
            raise ToolError(e.__context__)
        else:
            raise ToolError(e)
